//
//  YSXLog.m
//  YSXCrash
//
//  Created by Felix on 2018/8/24.
//  Copyright © 2018年 Felix. All rights reserved.
//

#import "VclLog.h"
#include <CoreFoundation/CoreFoundation.h>

/** The buffer size to use when writing log entries.
 *
 * If this value is > 0, any log entries that expand beyond this length will
 * be truncated.
 * If this value = 0, the logging system will dynamically allocate memory
 * and never truncate. However, the log functions won't be async-safe.
 *
 * Unless you're logging from within signal handlers, it's safe to set it to 0.
 */
#ifndef YSXLOGGER_CBufferSize
    #define YSXLOGGER_CBufferSize 1024
#endif

#pragma mark - C Function declaration

static inline void writeFmtToLog(const char* fmt, ...);
static inline void writeFmtArgsToLog(const char* fmt, va_list args);
static void writeToLog(const char* const str);
static inline void flushLog(void);
static inline const char* lastPathEntry(const char* const path);

#pragma mark - C Function implementation

void i_ysxlog_logC(const char *file, int line, const char *function, NSString *fmt, ...) {
    NSDateFormatter *formatter = [[NSDateFormatter alloc] init];
    formatter.dateFormat = @"yyyy-MM-dd HH:mm:ss.SSSSSS";
    NSString *dateString = [formatter stringFromDate:[NSDate date]];
    writeFmtToLog("%s %s (%u): %s: ", dateString.UTF8String, lastPathEntry(file), line, function);
    va_list args;
    va_start(args, fmt);
    writeFmtArgsToLog(fmt.UTF8String, args);
    va_end(args);
    writeToLog("\n");
    flushLog();
}

static inline void writeFmtToLog(const char* fmt, ...) {
    va_list args;
    va_start(args, fmt);
    writeFmtArgsToLog(fmt, args);
    va_end(args);
}

static inline void writeFmtArgsToLog(const char* fmt, va_list args) {
    if(__builtin_expect((fmt == NULL), 0)) {
        writeToLog("(null)");
    }
    else {
        char buffer[YSXLOGGER_CBufferSize];
        vsnprintf(buffer, sizeof(buffer), fmt, args);
        writeToLog(buffer);
    }
}

/** The file descriptor where log entries get written. */
static int g_fd = -1;

static void writeToLog(const char* const str) {
    if(g_fd >= 0) {
        int bytesToWrite = (int)strlen(str);
        const char* pos = str;
        while(bytesToWrite > 0) {
            int bytesWritten = (int)write(g_fd, pos, (unsigned)bytesToWrite);
            if(__builtin_expect((bytesWritten == -1), 0)) {
                break;
            }
            bytesToWrite -= bytesWritten;
            pos += bytesWritten;
        }
    }
    write(STDOUT_FILENO, str, strlen(str));
}

static inline void flushLog(void) {
    // Nothing to do.
}

static inline const char* lastPathEntry(const char* const path) {
    const char* lastFile = strrchr(path, '/');
    return lastFile == 0 ? path : lastFile + 1;
}

// ===========================================================================
#pragma mark - Objective-C -
// ===========================================================================

void i_ysxlog_logObjCBasic(CFStringRef fmt, ...) {
    if(fmt == NULL) {
        writeToLog("(null)");
        return;
    }
    
    va_list args;
    va_start(args, fmt);
    CFStringRef entry = CFStringCreateWithFormatAndArguments(NULL, NULL, fmt, args);
    va_end(args);
    
    int bufferLength = (int)CFStringGetLength(entry) * 4 + 1;
    char* stringBuffer = malloc((unsigned)bufferLength);
    if(CFStringGetCString(entry, stringBuffer, (CFIndex)bufferLength, kCFStringEncodingUTF8)) {
        NSDateFormatter *formatter = [[NSDateFormatter alloc] init];
        formatter.dateFormat = @"yyyy-MM-dd HH:mm:ss.SSSSSS ";
        NSString *dateString = [formatter stringFromDate:[NSDate date]];
        writeToLog(dateString.UTF8String);
        writeToLog(stringBuffer);
    }
    else {
        writeToLog("Could not convert log string to UTF-8. No logging performed.");
    }
    writeToLog("\n");
    
    free(stringBuffer);
    CFRelease(entry);
}

void i_ysxlog_logObjC(const char* const level, const char* const file, const int line, const char* const function, NSString *fmt, ...) {
    CFStringRef logFmt = NULL;
    if(fmt == NULL) {
        logFmt = CFStringCreateWithCString(NULL, "%s: %s (%u): %s: (null)", kCFStringEncodingUTF8);
        i_ysxlog_logObjCBasic(logFmt, level, lastPathEntry(file), line, function);
    }
    else {
        va_list args;
        va_start(args, fmt);
        CFStringRef entry = CFStringCreateWithFormatAndArguments(NULL, NULL, (__bridge CFStringRef)fmt, args);
        va_end(args);
        
        logFmt = CFStringCreateWithCString(NULL, "%s: %s (%u): %s: %@", kCFStringEncodingUTF8);
        i_ysxlog_logObjCBasic(logFmt, level, lastPathEntry(file), line, function, entry);
        
        CFRelease(entry);
    }
    CFRelease(logFmt);
}
