//
//  VclSystemInfo.m
//  VclCrash
//
//  Created by Felix on 2018/8/15.
//  Copyright © 2018年 Felix. All rights reserved.
//

#import "VclSystemInfo.h"

#import <UIKit/UIKit.h>
#import <sys/sysctl.h>
#include <mach-o/dyld.h>
#include <mach-o/nlist.h>
#include <mach-o/arch.h>
#import <CommonCrypto/CommonDigest.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <./time.h>

typedef struct {
    const char* systemName;
    const char* systemVersion;
    const char* machine;
    const char* model;
    const char* kernelVersion;
    const char* osVersion;
    const char* bootTime;
    const char* appStartTime;
    const char* executablePath;
    const char* executableName;
    const char* bundleID;
    const char* bundleName;
    const char* bundleVersion;
    const char* bundleShortVersion;
    const char* appID;
    const char* cpuArchitecture;
    const char* timezone;
    const char* processName;
    const char* deviceAppHash;
    const char* buildType;
    
    int cpuType;
    int cpuSubType;
    int binaryCPUType;
    int processID;
    int parentProcessID;
    int binaryCPUSubType;
    
    bool isJailbroken;
    
    uint64_t storageSize;
    uint64_t memorySize;
} VclSystemData;

#pragma mark - static method

@interface VclSystemInfo () {
    NSData *systemInfoData;
    VclSystemData systemData;
}

@end

static VclSystemInfo *_instance = nil;

@implementation VclSystemInfo

@synthesize systemInfoData = _systemInfoData;

+ (instancetype)sharedInstance {
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        _instance = [[VclSystemInfo alloc] init];
    });
    return _instance;
}

+ (instancetype)allocWithZone:(struct _NSZone *)zone {
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        _instance = [super allocWithZone:zone];
    });
    return _instance;
}

- (id)copyWithZone:(NSZone *)zone {
    return _instance;
}

- (instancetype)init {
    if (self = [super init]) {
        [self initialized];
    }
    return self;
}

- (void)initialized {
    NSBundle *mainBundle = [NSBundle mainBundle];
    NSDictionary *infoDict = [mainBundle infoDictionary];
    const struct mach_header* header = _dyld_get_image_header(0);
    
    systemData.systemName = cString([UIDevice currentDevice].systemName);
    systemData.systemVersion = cString([UIDevice currentDevice].systemVersion);
    systemData.machine = stringSysctl("hw.machine");
    systemData.model = stringSysctl("hw.model");
    systemData.kernelVersion = stringSysctl("kern.version");
    systemData.osVersion = stringSysctl("kern.osversion");
    systemData.isJailbroken = isJailbroken();
    systemData.bootTime = dateSysctl("kern.boottime");
    systemData.appStartTime = dateString(time(NULL));
    systemData.executablePath = cString(getExecutablePath());
    systemData.executableName = cString(infoDict[@"CFBundleExecutable"]);
    systemData.bundleID = cString(infoDict[@"CFBundleIdentifier"]);
    systemData.bundleName = cString(infoDict[@"CFBundleName"]);
    systemData.bundleVersion = cString(infoDict[@"CFBundleVersion"]);
    systemData.bundleShortVersion = cString(infoDict[@"CFBundleShortVersionString"]);
    systemData.appID = getAppUUID();
    systemData.cpuArchitecture = getCurrentCPUArch();
    systemData.cpuType = sysctl_int32ForName("hw.cputype");
    systemData.cpuSubType = sysctl_int32ForName("hw.cpusubtype");
    systemData.binaryCPUType = header->cputype;
    systemData.binaryCPUSubType = header->cpusubtype;
    systemData.timezone = cString([NSTimeZone localTimeZone].abbreviation);
    systemData.processName = cString([NSProcessInfo processInfo].processName);
    systemData.processID = [NSProcessInfo processInfo].processIdentifier;
    systemData.parentProcessID = getppid();
    systemData.deviceAppHash = getDeviceAndAppHash();
    systemData.buildType = getBuildType();
    systemData.storageSize = getStorageSize();
    systemData.memorySize = sysctl_uint64ForName("hw.memsize");
}

- (NSData *)systemInfoData {
    NSDateFormatter *dateFormatter = [[NSDateFormatter alloc] init];
    [dateFormatter setDateFormat:@"yyyy-MM-dd HH:mm:ss.sss"];
    NSMutableString *systemInfo = [[NSMutableString alloc] init];
    [systemInfo appendFormat:@"Incident Identifier: \t%s\r\n", systemData.appID ? systemData.appID : NULL];
    [systemInfo appendFormat:@"CrashReporter Key: \t%s\r\n", systemData.deviceAppHash ? systemData.deviceAppHash : NULL];
    [systemInfo appendFormat:@"Hardware Model: \t%s\r\n", systemData.machine ? systemData.machine : NULL];
    [systemInfo appendFormat:@"Process: \t\t%s [%d]\r\n", systemData.bundleName ? systemData.bundleName : NULL, systemData.processID];
    [systemInfo appendFormat:@"Path: \t\t\t%s\r\n", systemData.executablePath ? systemData.executablePath : NULL];
    [systemInfo appendFormat:@"Identifier: \t\t%s\r\n", systemData.bundleID ? systemData.bundleID : NULL];
    [systemInfo appendFormat:@"Version: \t\t%s (%s)\r\n", systemData.bundleShortVersion ? systemData.bundleShortVersion : NULL, systemData.bundleVersion ? systemData.bundleVersion : NULL];
    [systemInfo appendFormat:@"Code Type: \t\t%s\r\n", systemData.cpuArchitecture ? systemData.cpuArchitecture : NULL];
    [systemInfo appendFormat:@"isJailbroken: \t\t%@\r\n", systemData.isJailbroken ? @"YES" : @"NO"];
    [systemInfo appendFormat:@"Parent Process: \t? [%d]\r\n", systemData.parentProcessID];
    [systemInfo appendString:@"\r\n"];
    [systemInfo appendFormat:@"Date/Time: \t\t%@\r\n", [NSString stringWithFormat:@"%@ +%s", [dateFormatter stringFromDate:[NSDate date]], systemData.timezone]];
    [systemInfo appendFormat:@"OS Version: \t\t%s %s (%s)\r\n", systemData.systemName, systemData.systemVersion, systemData.osVersion];
    [systemInfo appendFormat:@"Build Type: \t\t%s", systemData.buildType];
    [systemInfo appendString:@"\r\n\r\n"];
    
    return [systemInfo dataUsingEncoding:NSUTF8StringEncoding];
}


const char* cString(NSString* str) {
    return str == NULL ? NULL : strdup(str.UTF8String);
}


int sysctl_stringForName(const char* const  name,
                         char* const value,
                         const int maxSize) {
    size_t size = value == NULL ? 0 : (size_t)maxSize;
    
    sysctlbyname(name, value, &size, NULL, 0);
    
    return (int)size;
}

static const char* stringSysctl(const char* name) {
    int size = (int)sysctl_stringForName(name, NULL, 0);
    if(size <= 0)
    {
        return NULL;
    }
    
    char* value = malloc((size_t)size);
    if(sysctl_stringForName(name, value, size) <= 0)
    {
        free(value);
        return NULL;
    }
    
    return value;
}

uint32_t dl_imageNamed(const char* const imageName, bool exactMatch) {
    if(imageName != NULL)
    {
        const uint32_t imageCount = _dyld_image_count();
        
        for(uint32_t iImg = 0; iImg < imageCount; iImg++)
        {
            const char* name = _dyld_get_image_name(iImg);
            if(exactMatch)
            {
                if(strcmp(name, imageName) == 0)
                {
                    return iImg;
                }
            }
            else
            {
                if(strstr(name, imageName) != NULL)
                {
                    return iImg;
                }
            }
        }
    }
    return UINT32_MAX;
}

static bool isJailbroken() {
    return dl_imageNamed("MobileSubstrate", false) != UINT32_MAX;
}

struct timeval sysctl_timevalForName(const char* const name) {
    struct timeval value = {0};
    size_t size = sizeof(value);
    
    if(0 != sysctlbyname(name, &value, &size, NULL, 0))
    {
        HLog(@"Could not get timeval value for %s: %s", name, strerror(errno));
    }
    
    return value;
}

void date_utcStringFromTimestamp(time_t timestamp, char* buffer21Chars) {
    struct tm result = {0};
    gmtime_r(&timestamp, &result);
    snprintf(buffer21Chars, 21, "%04d-%02d-%02dT%02d:%02d:%02dZ",
             result.tm_year + 1900,
             result.tm_mon+1,
             result.tm_mday,
             result.tm_hour,
             result.tm_min,
             result.tm_sec);
}

static const char* dateString(time_t date) {
    char* buffer = malloc(21);
    date_utcStringFromTimestamp(date, buffer);
    return buffer;
}

static const char* dateSysctl(const char* name) {
    struct timeval value = sysctl_timevalForName(name);
    return dateString(value.tv_sec);
}

/** Get this application's executable path.
 *
 * @return Executable path.
 */
static NSString* getExecutablePath() {
    NSBundle* mainBundle = [NSBundle mainBundle];
    NSDictionary* infoDict = [mainBundle infoDictionary];
    NSString* bundlePath = [mainBundle bundlePath];
    NSString* executableName = infoDict[@"CFBundleExecutable"];
    return [bundlePath stringByAppendingPathComponent:executableName];
}

static uintptr_t firstCmdAfterHeader(const struct mach_header* const header) {
    switch(header->magic)
    {
        case MH_MAGIC:
        case MH_CIGAM:
            return (uintptr_t)(header + 1);
        case MH_MAGIC_64:
        case MH_CIGAM_64:
            return (uintptr_t)(((struct mach_header_64*)header) + 1);
        default:
            // Header is corrupt
            return 0;
    }
}

const uint8_t* dl_imageUUID(const char* const imageName, bool exactMatch) {
    if(imageName != NULL)
    {
        const uint32_t iImg = dl_imageNamed(imageName, exactMatch);
        if(iImg != UINT32_MAX)
        {
            const struct mach_header* header = _dyld_get_image_header(iImg);
            if(header != NULL)
            {
                uintptr_t cmdPtr = firstCmdAfterHeader(header);
                if(cmdPtr != 0)
                {
                    for(uint32_t iCmd = 0;iCmd < header->ncmds; iCmd++)
                    {
                        const struct load_command* loadCmd = (struct load_command*)cmdPtr;
                        if(loadCmd->cmd == LC_UUID)
                        {
                            struct uuid_command* uuidCmd = (struct uuid_command*)cmdPtr;
                            return uuidCmd->uuid;
                        }
                        cmdPtr += loadCmd->cmdsize;
                    }
                }
            }
        }
    }
    return NULL;
}

static const char* uuidBytesToString(const uint8_t* uuidBytes) {
    CFUUIDRef uuidRef = CFUUIDCreateFromUUIDBytes(NULL, *((CFUUIDBytes*)uuidBytes));
    NSString* str = (__bridge_transfer NSString*)CFUUIDCreateString(NULL, uuidRef);
    CFRelease(uuidRef);
    
    return cString(str);
}

/** Get this application's UUID.
 *
 * @return The UUID.
 */
static const char* getAppUUID() {
    const char* result = nil;
    
    NSString* exePath = getExecutablePath();
    
    if(exePath != nil)
    {
        const uint8_t* uuidBytes = dl_imageUUID(exePath.UTF8String, true);
        if(uuidBytes == NULL)
        {
            // OSX app image path is a lie.
            uuidBytes = dl_imageUUID(exePath.lastPathComponent.UTF8String, false);
        }
        if(uuidBytes != NULL)
        {
            result = uuidBytesToString(uuidBytes);
        }
    }
    
    return result;
}

int32_t sysctl_int32ForName(const char* const name) {
    int32_t value = 0;
    size_t size = sizeof(value);
    
    sysctlbyname(name, &value, &size, NULL, 0);
    
    return value;
}

const char* cpu_currentArch(void) {
    const NXArchInfo* archInfo = NXGetLocalArchInfo();
    return archInfo == NULL ? NULL : archInfo->name;
}

static const char* getCPUArchForCPUType(cpu_type_t cpuType, cpu_subtype_t subType) {
    switch(cpuType)
    {
        case CPU_TYPE_ARM:
        {
            switch (subType)
            {
                case CPU_SUBTYPE_ARM_V6:
                    return "armv6";
                case CPU_SUBTYPE_ARM_V7:
                    return "armv7";
                case CPU_SUBTYPE_ARM_V7F:
                    return "armv7f";
                case CPU_SUBTYPE_ARM_V7K:
                    return "armv7k";
#ifdef CPU_SUBTYPE_ARM_V7S
                case CPU_SUBTYPE_ARM_V7S:
                    return "armv7s";
#endif
            }
            break;
        }
        case CPU_TYPE_X86:
            return "x86";
        case CPU_TYPE_X86_64:
            return "x86_64";
    }
    
    return NULL;
}

static const char* getCurrentCPUArch() {
    const char* result = getCPUArchForCPUType(sysctl_int32ForName("hw.cputype"),
                                              sysctl_int32ForName("hw.cpusubtype"));
    
    if(result == NULL)
    {
        result = cpu_currentArch();
    }
    return result;
}

static NSString* nsstringSysctl(NSString* name) {
    NSString* str = nil;
    int size = (int)sysctl_stringForName(name.UTF8String, NULL, 0);
    
    if(size <= 0)
    {
        return @"";
    }
    
    NSMutableData* value = [NSMutableData dataWithLength:(unsigned)size];
    
    if(sysctl_stringForName(name.UTF8String, value.mutableBytes, size) != 0)
    {
        str = [NSString stringWithCString:value.mutableBytes encoding:NSUTF8StringEncoding];
    }
    
    return str;
}

bool sysctl_getMacAddress(const char* const name,
                          char* const macAddressBuffer) {
    int mib[6] =
    {
        CTL_NET,
        AF_ROUTE,
        0,
        AF_LINK,
        NET_RT_IFLIST,
        (int)if_nametoindex(name)
    };
    if(mib[5] == 0)
    {
        HLog(@"Could not get interface index for %s: %s", name, strerror(errno));
        return false;
    }
    
    size_t length;
    if(sysctl(mib, 6, NULL, &length, NULL, 0) != 0)
    {
        HLog(@"Could not get interface data for %s: %s", name, strerror(errno));
        return false;
    }
    
    void* ifBuffer = malloc(length);
    if(ifBuffer == NULL)
    {
        HLog(@"Out of memory");
        return false;
    }
    
    if(sysctl(mib, 6, ifBuffer, &length, NULL, 0) != 0)
    {
        HLog(@"Could not get interface data for %s: %s", name, strerror(errno));
        free(ifBuffer);
        return false;
    }
    
    struct if_msghdr* msgHdr = (struct if_msghdr*) ifBuffer;
    struct sockaddr_dl* sockaddr = (struct sockaddr_dl*) &msgHdr[1];
    memcpy(macAddressBuffer, LLADDR(sockaddr), 6);
    
    free(ifBuffer);
    
    return true;
}

static const char* getDeviceAndAppHash() {
    NSMutableData* data = nil;
    
    if([[UIDevice currentDevice] respondsToSelector:@selector(identifierForVendor)])
    {
        data = [NSMutableData dataWithLength:16];
        [[UIDevice currentDevice].identifierForVendor getUUIDBytes:data.mutableBytes];
    }
    else
        
    {
        data = [NSMutableData dataWithLength:6];
        sysctl_getMacAddress("en0", [data mutableBytes]);
    }
    
    // Append some device-specific data.
    [data appendData:(NSData* _Nonnull )[nsstringSysctl(@"hw.machine") dataUsingEncoding:NSUTF8StringEncoding]];
    [data appendData:(NSData* _Nonnull )[nsstringSysctl(@"hw.model") dataUsingEncoding:NSUTF8StringEncoding]];
    const char* cpuArch = getCurrentCPUArch();
    [data appendBytes:cpuArch length:strlen(cpuArch)];
    
    // Append the bundle ID.
    NSData* bundleID = [[[NSBundle mainBundle] bundleIdentifier] dataUsingEncoding:NSUTF8StringEncoding];
    if(bundleID != nil)
    {
        [data appendData:bundleID];
    }
    
    // SHA the whole thing.
    uint8_t sha[CC_SHA1_DIGEST_LENGTH];
    CC_SHA1([data bytes], (CC_LONG)[data length], sha);
    
    NSMutableString* hash = [NSMutableString string];
    for(unsigned i = 0; i < sizeof(sha); i++)
    {
        [hash appendFormat:@"%02x", sha[i]];
    }
    
    return cString(hash);
}

static uint64_t getStorageSize() {
    NSNumber* storageSize = [[[NSFileManager defaultManager] attributesOfFileSystemForPath:NSHomeDirectory() error:nil] objectForKey:NSFileSystemSize];
    return storageSize.unsignedLongLongValue;
}

uint64_t sysctl_uint64ForName(const char* const name) {
    uint64_t value = 0;
    size_t size = sizeof(value);
    
    sysctlbyname(name, &value, &size, NULL, 0);
    
    return value;
}

static bool isSimulatorBuild() {
#if TARGET_OS_SIMULATOR
    return YES;
#else
    return NO;
#endif
}

static bool isDebugBuild() {
#ifdef DEBUG
    return YES;
#else
    return NO;
#endif
}

static NSString* getReceiptUrlPath() {
    NSString *path = [NSBundle mainBundle].appStoreReceiptURL.path;
    return path;
}

static bool hasAppStoreReceipt() {
    NSString* receiptPath = getReceiptUrlPath();
    if(receiptPath == nil)
    {
        return NO;
    }
    bool isAppStoreReceipt = [receiptPath.lastPathComponent isEqualToString:@"receipt"];
    bool receiptExists = [[NSFileManager defaultManager] fileExistsAtPath:receiptPath];
    
    return isAppStoreReceipt && receiptExists;
}

static const char* getBuildType() {
    if(isSimulatorBuild())
    {
        return "simulator";
    }
    if(isDebugBuild())
    {
        return "debug";
    }
    if(hasAppStoreReceipt())
    {
        return "app store";
    }
    return "unknown";
}


@end
